@@ -77,34 +77,18 @@ static void key_mnt_params_init(void)
 
 static void mnts_cleanup(void *arg)
 {
-	struct list_head *mnts = (struct list_head *) arg;
-	mnts_put_expire_list(mnts);
+	struct mnt_list *mnts = (struct mnt_list *) arg;
+	tree_free_mnt_tree(mnts);
+	return;
 }
 
-int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
+int do_umount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	struct mapent_cache *mc = me->mc;
 	char buf[MAX_ERR_BUF];
-	int ioctlfd = -1, rv, left, retries;
-	char key[PATH_MAX + 1];
-	struct mapent *tmp;
-	int opened = 0;
+	int ioctlfd, rv, left, retries;
 
-	if (me->len > PATH_MAX) {
-		error(ap->logopt, "path too long");
-		return 1;
-	}
-	strcpy(key, me->key);
-
-	cache_unlock(mc);
-	left = umount_multi(ap, key, 0);
-	cache_readlock(mc);
-	tmp = cache_lookup_distinct(mc, key);
-	if (tmp != me) {
-		error(ap->logopt, "key %s no longer in mapent cache", key);
-		return -1;
-	}
+	left = umount_multi(ap, me->key, 0);
 	if (left) {
 		warn(ap->logopt, "could not unmount %d dirs under %s",
 		     left, me->key);
@@ -112,20 +96,15 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 	}
 
 	if (me->ioctlfd != -1) {
-		if (ap->state == ST_READMAP &&
-		    is_mounted(me->key, MNTS_REAL)) {
+		if (tree_is_mounted(mnts, me->key, MNTS_REAL)) {
 			error(ap->logopt,
 			      "attempt to umount busy direct mount %s",
 			      me->key);
 			return 1;
 		}
 		ioctlfd = me->ioctlfd;
-	} else {
-		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
-		if (ioctlfd == -1)
-			return 1;
-		opened = 1;
-	}
+	} else
+		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
 
 	if (ioctlfd >= 0) {
 		unsigned int status = 1;
@@ -134,29 +113,22 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 		if (rv) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			error(ap->logopt, "ioctl failed: %s", estr);
-			/* The ioctl failed so this probably won't
-			 * work either but since we opened it here
-			 * try anyway. We should set these catatonic
-			 * too but ....
-			 */
-			if (opened)
-				ops->close(ap->logopt, ioctlfd);
 			return 1;
 		} else if (!status) {
 			if (ap->state != ST_SHUTDOWN_FORCE) {
 				error(ap->logopt,
 				      "ask umount returned busy for %s",
 				      me->key);
-				if (opened)
-					ops->close(ap->logopt, ioctlfd);
 				return 1;
 			} else {
 				me->ioctlfd = -1;
+				ops->catatonic(ap->logopt, ioctlfd);
 				ops->close(ap->logopt, ioctlfd);
 				goto force_umount;
 			}
 		}
 		me->ioctlfd = -1;
+		ops->catatonic(ap->logopt, ioctlfd);
 		ops->close(ap->logopt, ioctlfd);
 	} else {
 		error(ap->logopt,
@@ -168,7 +140,7 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 50000000};
+		struct timespec tm = {0, 200000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -186,11 +158,8 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 			warn(ap->logopt, "mount point %s is in use", me->key);
 			if (ap->state == ST_SHUTDOWN_FORCE)
 				goto force_umount;
-			else {
-				if (ap->state != ST_READMAP)
-					set_direct_mount_tree_catatonic(ap, me);
+			else
 				return 0;
-			}
 			break;
 		case ENOTDIR:
 			error(ap->logopt, "mount point is not a directory");
@@ -221,8 +190,20 @@ int umount_autofs_direct(struct autofs_point *ap)
 {
 	struct map_source *map;
 	struct mapent_cache *nc, *mc;
+	struct mnt_list *mnts;
 	struct mapent *me, *ne;
 
+	close(ap->state_pipe[0]);
+	close(ap->state_pipe[1]);
+	if (ap->pipefd >= 0)
+		close(ap->pipefd);
+	if (ap->kpipefd >= 0) {
+		close(ap->kpipefd);
+		ap->kpipefd = -1;
+	}
+
+	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
+	pthread_cleanup_push(mnts_cleanup, mnts);
 	nc = ap->entry->master->nc;
 	cache_readlock(nc);
 	pthread_cleanup_push(cache_lock_cleanup, nc);
@@ -231,70 +212,127 @@ int umount_autofs_direct(struct autofs_point *ap)
 		mc = map->mc;
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		cache_readlock(mc);
-restart:
 		me = cache_enumerate(mc, NULL);
 		while (me) {
-			int error;
-
 			ne = cache_lookup_distinct(nc, me->key);
 			if (ne && map->master_line > ne->age) {
 				me = cache_enumerate(mc, me);
 				continue;
 			}
 
-			/* The daemon is exiting so ...
-			 * If we get a fail here we must make our
-			 * best effort to set the direct mount trigger
-			 * catatonic regardless of the reason for the
-			 * failed umount.
-			 */
-			error = do_umount_autofs_direct(ap, me);
-			/* cache became invalid, restart */
-			if (error == -1)
-				goto restart;
-			if (!error)
-				goto done;
+			/* TODO: check return, locking me */
+			do_umount_autofs_direct(ap, mnts, me);
 
-			if (ap->state != ST_READMAP)
-				set_direct_mount_tree_catatonic(ap, me);
-done:
 			me = cache_enumerate(mc, me);
 		}
 		pthread_cleanup_pop(1);
 		map = map->next;
 	}
 	pthread_cleanup_pop(1);
-
-	if (ap->pipefd >= 0)
-		close(ap->pipefd);
-	if (ap->kpipefd >= 0) {
-		close(ap->kpipefd);
-		ap->kpipefd = -1;
-	}
+	pthread_cleanup_pop(1);
 
 	return 0;
 }
 
-int do_mount_autofs_direct(struct autofs_point *ap,
-			   struct mapent *me, time_t timeout)
+static int unlink_mount_tree(struct autofs_point *ap, struct list_head *list)
+{
+	struct list_head *p;
+	int rv, ret;
+	pid_t pgrp = getpgrp();
+	char spgrp[20];
+
+	sprintf(spgrp, "pgrp=%d", pgrp);
+
+	ret = 1;
+	list_for_each(p, list) {
+		struct mnt_list *mnt;
+
+		mnt = list_entry(p, struct mnt_list, list);
+
+		if (strstr(mnt->opts, spgrp))
+			continue;
+
+		if (strcmp(mnt->fs_type, "autofs"))
+			rv = spawn_umount(ap->logopt, "-l", mnt->path, NULL);
+		else
+			rv = umount2(mnt->path, MNT_DETACH);
+		if (rv == -1) {
+			debug(ap->logopt,
+			      "can't unlink %s from mount tree", mnt->path);
+
+			switch (errno) {
+			case EINVAL:
+				warn(ap->logopt,
+				      "bad superblock or not mounted");
+				break;
+
+			case ENOENT:
+			case EFAULT:
+				ret = 0;
+				warn(ap->logopt, "bad path for mount");
+				break;
+			}
+		}
+	}
+	return ret;
+}
+
+static int unlink_active_mounts(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me)
+{
+	struct ioctl_ops *ops = get_ioctl_ops();
+	struct list_head list;
+
+	INIT_LIST_HEAD(&list);
+
+	if (tree_get_mnt_list(mnts, &list, me->key, 1)) {
+		if (ap->state == ST_READMAP) {
+			time_t tout = ap->exp_timeout;
+			int save_ioctlfd, ioctlfd;
+
+			save_ioctlfd = ioctlfd = me->ioctlfd;
+
+			if (ioctlfd == -1)
+				ops->open(ap->logopt,
+					  &ioctlfd, me->dev, me->key);
+
+			if (ioctlfd < 0) {
+				error(ap->logopt,
+				     "failed to create ioctl fd for %s",
+				     me->key);
+				return 0;
+			}
+
+			ops->timeout(ap->logopt, ioctlfd, &tout);
+
+			if (save_ioctlfd == -1)
+				ops->close(ap->logopt, ioctlfd);
+
+			return 0;
+		}
+	}
+
+	if (!unlink_mount_tree(ap, &list)) {
+		debug(ap->logopt,
+		      "already mounted as other than autofs "
+		      "or failed to unlink entry in tree");
+		return 0;
+	}
+
+	return 1;
+}
+
+int do_mount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me)
 {
 	const char *str_direct = mount_type_str(t_direct);
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct mnt_params *mp;
+	time_t timeout = ap->exp_timeout;
 	struct stat st;
 	int status, ret, ioctlfd;
 	const char *map_name;
-	time_t runfreq;
-	int err;
 
-	if (timeout) {
-		/* Calculate the expire run frequency */
-		runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
-		if (ap->exp_runfreq)
-			ap->exp_runfreq = min(ap->exp_runfreq, runfreq);
-		else
-			ap->exp_runfreq = runfreq;
-	}
+	/* Calculate the timeouts */
+	ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
 
 	if (ops->version && !do_force_unlink) {
 		ap->flags |= MOUNT_FLAG_REMOUNT;
@@ -305,47 +343,16 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 		if (ret == 0)
 			return -1;
 	} else {
-		/* I don't remember why this is here for the force
-		 * unlink case. I don't think it should be but I may
-		 * have done it for a reason so keep it for the unlink
-		 * and continue case but not for the unlink and exit
-		 * case.
+		/*
+		 * A return of 0 indicates we're re-reading the map.
+		 * A return of 1 indicates we successfully unlinked
+		 * the mount tree if there was one. A return of -1
+		 * inducates we failed to unlink the mount tree so
+		 * we have to return a failure.
 		 */
-		if (!(do_force_unlink & UNLINK_AND_EXIT) &&
-		    ap->state == ST_READMAP && is_mounted(me->key, MNTS_ALL)) {
-			time_t tout = get_exp_timeout(ap, me->source);
-			int save_ioctlfd, ioctlfd;
-
-			save_ioctlfd = ioctlfd = me->ioctlfd;
-
-			if (ioctlfd == -1)
-				ioctlfd = open_ioctlfd(ap, me->key, me->dev);
-
-			if (ioctlfd < 0) {
-				error(ap->logopt,
-				     "failed to create ioctl fd for %s",
-				     me->key);
-				return 0;
-			}
-
-			ops->timeout(ap->logopt, ioctlfd, tout);
-
-			if (save_ioctlfd == -1)
-				ops->close(ap->logopt, ioctlfd);
-
-			return 0;
-		}
-
-		ret = unlink_mount_tree(ap, me->key);
-		if (!ret) {
-			error(ap->logopt,
-			     "already mounted as other than autofs "
-			     "or failed to unlink entry in tree");
-			goto out_err;
-		}
-
-		if (do_force_unlink & UNLINK_AND_EXIT)
-			return -1;
+		ret = unlink_active_mounts(ap, mnts, me);
+		if (ret == -1 || ret == 0)
+			return ret;
 
 		if (me->ioctlfd != -1) {
 			error(ap->logopt, "active direct mount %s", me->key);
@@ -376,14 +383,13 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path,
-				ap->kpipefd, str_direct, ap->flags);
+		mp->options = make_options_string(ap->path, ap->kpipefd, str_direct);
 		if (!mp->options)
 			return 0;
 	}
 
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(me->key, mp_mode) < 0) {
+	if (mkdir_path(me->key, 0555) < 0) {
 		if (errno != EEXIST && errno != EROFS) {
 			crit(ap->logopt,
 			     "failed to create mount directory %s", me->key);
@@ -411,21 +417,16 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 		      "failed to stat direct mount trigger %s", me->key);
 		goto out_umount;
 	}
-	me->dev = st.st_dev;
-	me->ino = st.st_ino;
 
-	if (ap->mode && (err = chmod(me->key, ap->mode)))
-		warn(ap->logopt, "failed to change mode of %s", me->key);
-
-	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
+	ops->open(ap->logopt, &ioctlfd, st.st_dev, me->key);
 	if (ioctlfd < 0) {
 		crit(ap->logopt, "failed to create ioctl fd for %s", me->key);
 		goto out_umount;
 	}
 
-	ops->timeout(ap->logopt, ioctlfd, timeout);
-	notify_mount_result(ap, me->key, timeout, str_direct);
-	cache_set_ino_index(me->mc, me);
+	ops->timeout(ap->logopt, ioctlfd, &timeout);
+	notify_mount_result(ap, me->key, str_direct);
+	cache_set_ino_index(me->mc, me->key, st.st_dev, st.st_ino);
 	ops->close(ap->logopt, ioctlfd);
 
 	debug(ap->logopt, "mounted trigger %s", me->key);
@@ -447,7 +448,8 @@ int mount_autofs_direct(struct autofs_point *ap)
 	struct map_source *map;
 	struct mapent_cache *nc, *mc;
 	struct mapent *me, *ne, *nested;
-	time_t now = monotonic_time(NULL);
+	struct mnt_list *mnts;
+	time_t now = time(NULL);
 
 	if (strcmp(ap->path, "/-")) {
 		error(ap->logopt, "expected direct map, exiting");
@@ -462,12 +464,15 @@ int mount_autofs_direct(struct autofs_point *ap)
 		return -1;
 	}
 
+	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
+	pthread_cleanup_push(mnts_cleanup, mnts);
 	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 	master_source_readlock(ap->entry);
 	nc = ap->entry->master->nc;
+	cache_readlock(nc);
+	pthread_cleanup_push(cache_lock_cleanup, nc);
 	map = ap->entry->maps;
 	while (map) {
-		time_t timeout;
 		/*
 		 * Only consider map sources that have been read since
 		 * the map entry was last updated.
@@ -478,38 +483,32 @@ int mount_autofs_direct(struct autofs_point *ap)
 		}
 
 		mc = map->mc;
-		timeout = get_exp_timeout(ap, map);
 		cache_readlock(mc);
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
-			cache_writelock(nc);
 			ne = cache_lookup_distinct(nc, me->key);
 			if (ne) {
-				unsigned int ne_age = ne->age;
-
-				cache_unlock(nc);
-				if (map->master_line < ne_age) {
+				if (map->master_line < ne->age) {
 					/* TODO: check return, locking me */
-					do_mount_autofs_direct(ap, me, timeout);
+					do_mount_autofs_direct(ap, mnts, me);
 				}
 				me = cache_enumerate(mc, me);
 				continue;
 			}
 
 			nested = cache_partial_match(nc, me->key);
-			if (!nested)
-				cache_unlock(nc);
-			else {
-				cache_delete(nc, nested->key);
-				cache_unlock(nc);
+			if (nested) {
 				error(ap->logopt,
 				   "removing invalid nested null entry %s",
 				   nested->key);
+				nested = cache_partial_match(nc, me->key);
+				if (nested)
+					cache_delete(nc, nested->key);
 			}
 
 			/* TODO: check return, locking me */
-			do_mount_autofs_direct(ap, me, timeout);
+			do_mount_autofs_direct(ap, mnts, me);
 
 			me = cache_enumerate(mc, me);
 		}
@@ -517,6 +516,8 @@ int mount_autofs_direct(struct autofs_point *ap)
 		map = map->next;
 	}
 	pthread_cleanup_pop(1);
+	pthread_cleanup_pop(1);
+	pthread_cleanup_pop(1);
 
 	return 0;
 }
@@ -525,11 +526,10 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	char buf[MAX_ERR_BUF];
-	int ioctlfd = -1, rv = 1, retries;
-	int opened = 0;
+	int ioctlfd, rv = 1, retries;
 
 	if (me->ioctlfd != -1) {
-		if (is_mounted(me->key, MNTS_REAL)) {
+		if (is_mounted(_PATH_MOUNTED, me->key, MNTS_REAL)) {
 			error(ap->logopt,
 			      "attempt to umount busy offset %s", me->key);
 			return 1;
@@ -537,16 +537,13 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 		ioctlfd = me->ioctlfd;
 	} else {
 		/* offset isn't mounted, return success and try to recover */
-		if (!is_mounted(me->key, MNTS_AUTOFS)) {
+		if (!is_mounted(_PROC_MOUNTS, me->key, MNTS_AUTOFS)) {
 			debug(ap->logopt,
 			      "offset %s not mounted",
 			      me->key);
 			return 0;
 		}
-		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
-		if (ioctlfd == -1)
-			return 1;
-		opened = 1;
+		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
 	}
 
 	if (ioctlfd >= 0) {
@@ -556,8 +553,6 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 		if (rv) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			logerr("ioctl failed: %s", estr);
-			if (opened && ioctlfd != -1)
-				ops->close(ap->logopt, ioctlfd);
 			return 1;
 		} else if (!status) {
 			if (ap->state != ST_SHUTDOWN_FORCE) {
@@ -565,8 +560,6 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 					error(ap->logopt,
 					     "ask umount returned busy for %s",
 					     me->key);
-				if (opened && ioctlfd != -1)
-					ops->close(ap->logopt, ioctlfd);
 				return 1;
 			} else {
 				me->ioctlfd = -1;
@@ -598,7 +591,7 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 50000000};
+		struct timespec tm = {0, 200000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -630,43 +623,44 @@ force_umount:
 	} else
 		info(ap->logopt, "umounted offset mount %s", me->key);
 
+	if (!rv && me->flags & MOUNT_FLAG_DIR_CREATED) {
+		if  (rmdir(me->key) == -1) {
+			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			warn(ap->logopt, "failed to remove dir %s: %s",
+			     me->key, estr);
+		}
+	}
 	return rv;
 }
 
-int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
+int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *root, const char *offset)
 {
 	const char *str_offset = mount_type_str(t_offset);
 	struct ioctl_ops *ops = get_ioctl_ops();
 	char buf[MAX_ERR_BUF];
 	struct mnt_params *mp;
-	time_t timeout = get_exp_timeout(ap, me->source);
+	time_t timeout = ap->exp_timeout;
 	struct stat st;
 	int ioctlfd, status, ret;
 	const char *hosts_map_name = "-hosts";
 	const char *map_name = hosts_map_name;
 	const char *type;
-	struct mnt_list *mnt;
+	char mountpoint[PATH_MAX];
 
 	if (ops->version && ap->flags & MOUNT_FLAG_REMOUNT) {
 		ret = try_remount(ap, me, t_offset);
 		if (ret == 1)
 			return MOUNT_OFFSET_OK;
-		/* Offset mount not found, fall thru and try to mount it */
-		if (!(ret == -1 && errno == ENOENT))
-			return MOUNT_OFFSET_FAIL;
+		return MOUNT_OFFSET_FAIL;
 	} else {
-		if (is_mounted(me->key, MNTS_AUTOFS)) {
+/*
+		if (is_mounted(_PROC_MOUNTS, me->key, MNTS_AUTOFS)) {
 			if (ap->state != ST_READMAP)
 				warn(ap->logopt,
 				     "trigger %s already mounted", me->key);
-			mnt = mnts_add_mount(ap, me->key, MNTS_OFFSET);
-			if (!mnt)
-				error(ap->logopt,
-				      "failed to add offset mount %s to mounted list",
-				      me->key);
 			return MOUNT_OFFSET_OK;
 		}
-
+*/
 		if (me->ioctlfd != -1) {
 			error(ap->logopt, "active offset mount %s", me->key);
 			return MOUNT_OFFSET_FAIL;
@@ -696,14 +690,16 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path,
-				ap->kpipefd, str_offset, ap->flags);
+		mp->options = make_options_string(ap->path, ap->kpipefd, str_offset);
 		if (!mp->options)
 			return MOUNT_OFFSET_OK;
 	}
 
+	strcpy(mountpoint, root);
+	strcat(mountpoint, offset);
+
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(me->key, mp_mode) < 0) {
+	if (mkdir_path(mountpoint, 0555) < 0) {
 		if (errno == EEXIST) {
 			/*
 			 * If the mount point directory is a real mount
@@ -711,8 +707,8 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 			 * a mount that has been automatically mounted by
 			 * the kernel NFS client.
 			 */
-			if (!IS_MM_ROOT(me) &&
-			    is_mounted(me->key, MNTS_REAL))
+			if (me->multi != me &&
+			    is_mounted(_PROC_MOUNTS, mountpoint, MNTS_REAL))
 				return MOUNT_OFFSET_IGNORE;
 
 			/* 
@@ -732,13 +728,13 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			debug(ap->logopt,
 			     "can't create mount directory: %s, %s",
-			     me->key, estr);
+			     mountpoint, estr);
 			return MOUNT_OFFSET_FAIL;
 		} else {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			crit(ap->logopt,
 			     "failed to create mount directory: %s, %s",
-			     me->key, estr);
+			     mountpoint, estr);
 			return MOUNT_OFFSET_FAIL;
 		}
 	} else {
@@ -748,51 +744,50 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 
 	debug(ap->logopt,
 	      "calling mount -t autofs " SLOPPY " -o %s automount %s",
-	      mp->options, me->key);
+	      mp->options, mountpoint);
 
 	type = ap->entry->maps->type;
 	if (!type || strcmp(ap->entry->maps->type, "hosts"))
 		map_name = me->mc->map->argv[0];
 
-	ret = mount(map_name, me->key, "autofs", MS_MGC_VAL, mp->options);
+	ret = mount(map_name, mountpoint, "autofs", MS_MGC_VAL, mp->options);
 	if (ret) {
 		crit(ap->logopt,
 		     "failed to mount offset trigger %s at %s",
-		     me->key, me->key);
+		     me->key, mountpoint);
 		goto out_err;
 	}
 
-	ret = stat(me->key, &st);
+	ret = stat(mountpoint, &st);
 	if (ret == -1) {
-		int save_errno = errno;
-
 		error(ap->logopt,
-		     "failed to stat direct mount trigger %s", me->key);
-		if (save_errno != ENOENT)
-			goto out_umount;
-		goto out_err;
-	}
-	me->dev = st.st_dev;
-	me->ino = st.st_ino;
-
-	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
-	if (ioctlfd < 0)
+		     "failed to stat direct mount trigger %s", mountpoint);
 		goto out_umount;
+	}
 
-	ops->timeout(ap->logopt, ioctlfd, timeout);
-	cache_set_ino_index(me->mc, me);
-	notify_mount_result(ap, me->key, timeout, str_offset);
+	ops->open(ap->logopt, &ioctlfd, st.st_dev, mountpoint);
+	if (ioctlfd < 0) {
+		crit(ap->logopt, "failed to create ioctl fd for %s", mountpoint);
+		goto out_umount;
+	}
+
+	ops->timeout(ap->logopt, ioctlfd, &timeout);
+	cache_set_ino_index(me->mc, me->key, st.st_dev, st.st_ino);
+	if (ap->logopt & LOGOPT_DEBUG)
+		notify_mount_result(ap, mountpoint, str_offset);
+	else
+		notify_mount_result(ap, me->key, str_offset);
 	ops->close(ap->logopt, ioctlfd);
 
-	debug(ap->logopt, "mounted trigger %s", me->key);
+	debug(ap->logopt, "mounted trigger %s at %s", me->key, mountpoint);
 
 	return MOUNT_OFFSET_OK;
 
 out_umount:
-	umount(me->key);
+	umount(mountpoint);
 out_err:
-	if (stat(me->key, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
-		 rmdir_path(ap, me->key, st.st_dev);
+	if (stat(mountpoint, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
+		 rmdir_path(ap, mountpoint, st.st_dev);
 
 	return MOUNT_OFFSET_FAIL;
 }
@@ -800,13 +795,13 @@ out_err:
 void *expire_proc_direct(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	struct mnt_list *mnt;
-	LIST_HEAD(mnts);
+	struct mnt_list *mnts = NULL, *next;
+	struct list_head list, *p;
 	struct expire_args *ea;
 	struct expire_args ec;
 	struct autofs_point *ap;
 	struct mapent *me = NULL;
-	unsigned int how;
+	unsigned int now;
 	int ioctlfd, cur_state;
 	int status, ret, left;
 
@@ -817,7 +812,7 @@ void *expire_proc_direct(void *arg)
 		fatal(status);
 
 	ap = ec.ap = ea->ap;
-	how = ea->how;
+	now = ea->when;
 	ec.status = -1;
 
 	ea->signaled = 1;
@@ -833,24 +828,31 @@ void *expire_proc_direct(void *arg)
 
 	left = 0;
 
-	/* Get the list of real mounts and expire them if possible */
-	mnts_get_expire_list(&mnts, ap);
-	if (list_empty(&mnts))
-		goto done;
-	pthread_cleanup_push(mnts_cleanup, &mnts);
-	list_for_each_entry(mnt, &mnts, expire) {
+	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
+	pthread_cleanup_push(mnts_cleanup, mnts);
+
+	/* Get a list of mounts select real ones and expire them if possible */
+	INIT_LIST_HEAD(&list);
+	if (!tree_get_mnt_list(mnts, &list, "/", 0)) {
+		ec.status = 0;
+		return NULL;
+	}
+
+	list_for_each(p, &list) {
+		next = list_entry(p, struct mnt_list, list);
+
 		/*
 		 * All direct mounts must be present in the map
 		 * entry cache.
 		 */
 		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 		master_source_readlock(ap->entry);
-		me = lookup_source_mapent(ap, mnt->mp, LKP_DISTINCT);
+		me = lookup_source_mapent(ap, next->path, LKP_DISTINCT);
 		pthread_cleanup_pop(1);
 		if (!me)
 			continue;
 
-		if (mnt->flags & (MNTS_AUTOFS|MNTS_OFFSET)) {
+		if (!strcmp(next->fs_type, "autofs")) {
 			struct stat st;
 			int ioctlfd;
 
@@ -861,8 +863,8 @@ void *expire_proc_direct(void *arg)
 			 * one of them and pass on state change.
 			 */
 			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-			if (mnt->flags & MNTS_AUTOFS) {
-				master_notify_submount(ap, mnt->mp, ap->state);
+			if (strstr(next->opts, "indirect")) {
+				master_notify_submount(ap, next->path, ap->state);
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
@@ -872,9 +874,11 @@ void *expire_proc_direct(void *arg)
 				continue;
 			}
 
-			/* It's got a mount, just send the expire. */
-			if (is_mounted(me->key, MNTS_REAL))
-				goto cont;
+			/* It's got a mount, deal with in the outer loop */
+			if (tree_is_mounted(mnts, me->key, MNTS_REAL)) {
+				pthread_setcancelstate(cur_state, NULL);
+				continue;
+			}
 
 			/*
 			 * Maybe a manual umount, repair.
@@ -891,20 +895,19 @@ void *expire_proc_direct(void *arg)
 			cache_writelock(me->mc);
 			if (me->ioctlfd != -1 && 
 			    fstat(me->ioctlfd, &st) != -1 &&
-			    !count_mounts(ap, mnt->mp, st.st_dev)) {
+			    !count_mounts(ap->logopt, next->path, st.st_dev)) {
 				ops->close(ap->logopt, me->ioctlfd);
 				me->ioctlfd = -1;
 				cache_unlock(me->mc);
-				mnts_remove_mount(mnt->mp, MNTS_MOUNTED);
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
 			cache_unlock(me->mc);
-cont:
+
 			ioctlfd = me->ioctlfd;
 
-			ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
-			if (ret == 1) {
+			ret = ops->expire(ap->logopt, ioctlfd, next->path, now);
+			if (ret) {
 				left++;
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
@@ -926,19 +929,19 @@ cont:
 		if (ap->state == ST_EXPIRE || ap->state == ST_PRUNE)
 			pthread_testcancel();
 
-		debug(ap->logopt, "send expire to trigger %s", mnt->mp);
+		debug(ap->logopt, "send expire to trigger %s", next->path);
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
-		if (ret == 1)
+		ret = ops->expire(ap->logopt, ioctlfd, next->path, now);
+		if (ret)
 			left++;
 		pthread_setcancelstate(cur_state, NULL);
 	}
 	pthread_cleanup_pop(1);
 
 	if (left)
-		debug(ap->logopt, "%d remaining in %s", left, ap->path);
-done:
+		info(ap->logopt, "%d remaining in %s", left, ap->path);
+
 	ec.status = left;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -998,8 +1001,7 @@ static void *do_expire_direct(void *arg)
 		struct mapent *me;
 		cache_writelock(mt.mc);
 		me = cache_lookup_distinct(mt.mc, mt.name);
-		if (me)
-			me->ioctlfd = -1;
+		me->ioctlfd = -1;
 		cache_unlock(mt.mc);
 		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
 		ops->close(ap->logopt, mt.ioctlfd);
@@ -1021,6 +1023,7 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	char buf[MAX_ERR_BUF];
 	pthread_t thid;
 	struct timespec wait;
+	struct timeval now;
 	int status, state;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
@@ -1047,18 +1050,13 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 		map = map->next;
 	}
 
-	if (!me || me->len >= PATH_MAX) {
+	if (!me) {
 		/*
 		 * Shouldn't happen as we have been sent this following
 		 * successful thread creation and lookup.
 		 */
-		if (!me)
-			crit(ap->logopt, "can't find map entry for (%lu,%lu)",
-			    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
-		else {
-			cache_unlock(mc);
-			crit(ap->logopt, "lookup key is too long");
-		}
+		crit(ap->logopt, "can't find map entry for (%lu,%lu)",
+		    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
 		master_source_unlock(ap->entry);
 		pthread_setcancelstate(state, NULL);
 		return 1;
@@ -1067,9 +1065,9 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	/* Can't expire it if it isn't mounted */
 	if (me->ioctlfd == -1) {
 		int ioctlfd;
-
-		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
+		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
 		if (ioctlfd == -1) {
+			crit(ap->logopt, "can't open ioctlfd for %s", me->key);
 			cache_unlock(mc);
 			master_source_unlock(ap->entry);
 			pthread_setcancelstate(state, NULL);
@@ -1095,7 +1093,9 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 		return 1;
 	}
 
-	pending_cond_init(mt);
+	status = pthread_cond_init(&mt->cond, NULL);
+	if (status)
+		fatal(status);
 
 	status = pthread_mutex_init(&mt->mutex, NULL);
 	if (status)
@@ -1104,6 +1104,7 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	mt->ap = ap;
 	mt->ioctlfd = me->ioctlfd;
 	mt->mc = mc;
+	/* TODO: check length here */
 	strcpy(mt->name, me->key);
 	mt->dev = me->dev;
 	mt->type = NFY_EXPIRE;
@@ -1140,8 +1141,9 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 
 	mt->signaled = 0;
 	while (!mt->signaled) {
-		clock_gettime(CLOCK_MONOTONIC, &wait);
-		wait.tv_sec += 2;
+		gettimeofday(&now, NULL);
+		wait.tv_sec = now.tv_sec + 2;
+		wait.tv_nsec = now.tv_usec * 1000;
 		status = pthread_cond_timedwait(&mt->cond, &mt->mutex, &wait);
 		if (status && status != ETIMEDOUT)
 			fatal(status);
@@ -1155,18 +1157,12 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	return 0;
 }
 
-static void mount_send_status(void *arg)
+static void mount_send_fail(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt = arg;
 	struct autofs_point *ap = mt->ap;
-
-	if (mt->status)
-		ops->send_fail(ap->logopt, mt->ioctlfd,
-			       mt->wait_queue_token, mt->status);
-	else
-		ops->send_ready(ap->logopt,
-				mt->ioctlfd, mt->wait_queue_token);
+	ops->send_fail(ap->logopt, mt->ioctlfd, mt->wait_queue_token, -ENOENT);
 	ops->close(ap->logopt, mt->ioctlfd);
 }
 
@@ -1175,7 +1171,6 @@ static void *do_mount_direct(void *arg)
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *args, mt;
 	struct autofs_point *ap;
-	struct mapent *me;
 	struct stat st;
 	int status, state;
 
@@ -1187,8 +1182,6 @@ static void *do_mount_direct(void *arg)
 
 	ap = mt.ap;
 
-	set_thread_mount_request_log_id(&mt);
-
 	args->signaled = 1;
 	status = pthread_cond_signal(&args->cond);
 	if (status)
@@ -1196,23 +1189,28 @@ static void *do_mount_direct(void *arg)
 
 	pending_mutex_unlock(args);
 
-	mt.status = 0;
-	pthread_cleanup_push(mount_send_status, &mt);
+	pthread_cleanup_push(mount_send_fail, &mt);
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
-	if (defaults_get_mount_verbose()) {
-		pid_t ppid = log_pidinfo(ap, mt.pid, "requestor");
-		if (ppid > 0)
-			log_pidinfo(ap, ppid, "parent");
+	status = fstat(mt.ioctlfd, &st);
+	if (status == -1) {
+		error(ap->logopt,
+		      "can't stat direct mount trigger %s", mt.name);
+		ops->send_fail(ap->logopt,
+			       mt.ioctlfd, mt.wait_queue_token, -ENOENT);
+		ops->close(ap->logopt, mt.ioctlfd);
+		pthread_setcancelstate(state, NULL);
+		pthread_exit(NULL);
 	}
 
-	status = fstat(mt.ioctlfd, &st);
+	status = stat(mt.name, &st);
 	if (status != 0 || !S_ISDIR(st.st_mode) || st.st_dev != mt.dev) {
 		error(ap->logopt,
 		     "direct trigger not valid or already mounted %s",
 		     mt.name);
-		mt.status = -EINVAL;
+		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
+		ops->close(ap->logopt, mt.ioctlfd);
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
@@ -1230,17 +1228,14 @@ static void *do_mount_direct(void *arg)
 	 */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 	if (status) {
-		struct mnt_list *sbmnt;
+		struct mapent *me;
 		struct statfs fs;
 		unsigned int close_fd = 0;
-		unsigned int flags = MNTS_DIRECT|MNTS_MOUNTED;
 
-		sbmnt = mnts_find_submount(mt.name);
 		if (statfs(mt.name, &fs) == -1 ||
-		   (fs.f_type == AUTOFS_SUPER_MAGIC && !sbmnt))
+		   (fs.f_type == AUTOFS_SUPER_MAGIC &&
+		    !master_find_submount(ap, mt.name)))
 			close_fd = 1;
-		if (sbmnt)
-			mnts_put_mount(sbmnt);
 		cache_writelock(mt.mc);
 		if ((me = cache_lookup_distinct(mt.mc, mt.name))) {
 			/*
@@ -1248,38 +1243,22 @@ static void *do_mount_direct(void *arg)
 			 * for direct mount multi-mounts with no real mount at
 			 * their base so they will be expired.
 			 */
-			if (close_fd && IS_MM_ROOT(me))
+			if (close_fd && me == me->multi)
 				close_fd = 0;
 			if (!close_fd)
 				me->ioctlfd = mt.ioctlfd;
-			if (IS_MM(me) && !IS_MM_ROOT(me))
-				flags |= MNTS_OFFSET;
 		}
 		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
 		cache_unlock(mt.mc);
 		if (close_fd)
 			ops->close(ap->logopt, mt.ioctlfd);
-
 		info(ap->logopt, "mounted %s", mt.name);
-
-		mnts_set_mounted_mount(ap, mt.name, flags);
-
-		conditional_alarm_add(ap, ap->exp_runfreq);
 	} else {
 		/* TODO: get mount return status from lookup_nss_mount */
 		ops->send_fail(ap->logopt,
 			       mt.ioctlfd, mt.wait_queue_token, -ENOENT);
 		ops->close(ap->logopt, mt.ioctlfd);
 		info(ap->logopt, "failed to mount %s", mt.name);
-
-		/* If this is a multi-mount subtree mount failure
-		 * ensure the tree continues to expire.
-		 */
-		cache_readlock(mt.mc);
-		me = cache_lookup_distinct(mt.mc, mt.name);
-		if (me && IS_MM(me) && !IS_MM_ROOT(me))
-			conditional_alarm_add(ap, ap->exp_runfreq);
-		cache_unlock(mt.mc);
 	}
 	pthread_setcancelstate(state, NULL);
 
@@ -1299,9 +1278,8 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 	char buf[MAX_ERR_BUF];
 	int status = 0;
 	struct timespec wait;
-	int ioctlfd, state;
-	unsigned int kver_major = get_kver_major();
-	unsigned int kver_minor = get_kver_minor();
+	struct timeval now;
+	int ioctlfd, len, state;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
@@ -1314,16 +1292,8 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 	 * cache entry we will not be able to find the mapent. So
 	 * we must take the source writelock to ensure the parent
 	 * has mount is complete before we look for the entry.
-	 *
-	 * Since the vfs-automount kernel changes we can now block
-	 * on covered mounts during mount tree construction so a
-	 * write lock is no longer needed. So we now can handle a
-	 * wider class of recursively define mount lookups.
 	 */
-	if (kver_major > 5 || (kver_major == 5 && kver_minor > 1))
-		master_source_readlock(ap->entry);
-	else
-		master_source_writelock(ap->entry);
+	master_source_writelock(ap->entry);
 	map = ap->entry->maps;
 	while (map) {
 		/*
@@ -1336,7 +1306,7 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		}
 
 		mc = map->mc;
-		cache_readlock(mc);
+		cache_writelock(mc);
 		me = cache_lookup_ino(mc, pkt->dev, pkt->ino);
 		if (me)
 			break;
@@ -1352,7 +1322,6 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		logerr("can't find map entry for (%lu,%lu)",
 		    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
 		master_source_unlock(ap->entry);
-		master_mutex_unlock();
 		pthread_setcancelstate(state, NULL);
 		return 1;
 	}
@@ -1362,8 +1331,8 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		close(me->ioctlfd);
 		me->ioctlfd = -1;
 	}
+	ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
 
-	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
 	if (ioctlfd == -1) {
 		cache_unlock(mc);
 		master_source_unlock(ap->entry);
@@ -1386,22 +1355,11 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		master_source_unlock(ap->entry);
 		master_mutex_unlock();
 		pthread_setcancelstate(state, NULL);
-		return 0;
+		return 1;
 	}
 
-	/* Check if we recorded a mount fail for this key */
-	if (me->status >= monotonic_time(NULL)) {
-		ops->send_fail(ap->logopt,
-			       ioctlfd, pkt->wait_queue_token, -ENOENT);
-		ops->close(ap->logopt, ioctlfd);
-		cache_unlock(mc);
-		master_source_unlock(ap->entry);
-		master_mutex_unlock();
-		pthread_setcancelstate(state, NULL);
-		return 0;
-	}
-
-	if (me->len >= PATH_MAX) {
+	len = strlen(me->key);
+	if (len >= PATH_MAX) {
 		error(ap->logopt, "direct mount path too long %s", me->key);
 		ops->send_fail(ap->logopt,
 			       ioctlfd, pkt->wait_queue_token, -ENAMETOOLONG);
@@ -1410,7 +1368,7 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		master_source_unlock(ap->entry);
 		master_mutex_unlock();
 		pthread_setcancelstate(state, NULL);
-		return 0;
+		return 1;
 	}
 
 	mt = malloc(sizeof(struct pending_args));
@@ -1424,11 +1382,13 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		master_source_unlock(ap->entry);
 		master_mutex_unlock();
 		pthread_setcancelstate(state, NULL);
-		return 0;
+		return 1;
 	}
 	memset(mt, 0, sizeof(struct pending_args));
 
-	pending_cond_init(mt);
+	status = pthread_cond_init(&mt->cond, NULL);
+	if (status)
+		fatal(status);
 
 	status = pthread_mutex_init(&mt->mutex, NULL);
 	if (status)
@@ -1440,12 +1400,11 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 	mt->ioctlfd = ioctlfd;
 	mt->mc = mc;
 	strcpy(mt->name, me->key);
-	mt->len = me->len;
+	mt->len = len;
 	mt->dev = me->dev;
 	mt->type = NFY_MOUNT;
 	mt->uid = pkt->uid;
 	mt->gid = pkt->gid;
-	mt->pid = pkt->pid;
 	mt->wait_queue_token = pkt->wait_queue_token;
 
 	status = pthread_create(&thid, &th_attr_detached, do_mount_direct, mt);
@@ -1478,8 +1437,9 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 
 	mt->signaled = 0;
 	while (!mt->signaled) {
-		clock_gettime(CLOCK_MONOTONIC, &wait);
-		wait.tv_sec += 2;
+		gettimeofday(&now, NULL);
+		wait.tv_sec = now.tv_sec + 2;
+		wait.tv_nsec = now.tv_usec * 1000;
 		status = pthread_cond_timedwait(&mt->cond, &mt->mutex, &wait);
 		if (status && status != ETIMEDOUT)
 			fatal(status);
