@@ -22,7 +22,6 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <unistd.h>
 #include "automount.h"
 #include "nsswitch.h"
 
@@ -158,6 +157,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			result = do_read_master(master, "file", age);
 		}
 
+		if (result == NSS_STATUS_UNAVAIL)
+			master->read_fail = 1;
+
 		return !result;
 	} else {
 		char *name = master->name;
@@ -195,6 +197,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 				result = do_read_master(master, source, age);
 				master->name = name;
 
+				if (result == NSS_STATUS_UNAVAIL)
+					master->read_fail = 1;
+
 				return !result;
 			}
 		}
@@ -249,6 +254,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			continue;
 		}
 
+		if (result == NSS_STATUS_UNAVAIL)
+			master->read_fail = 1;
+
 		status = check_nss_result(this, result);
 		if (status >= 0) {
 			free_sources(&nsslist);
@@ -279,17 +287,6 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 	map->lookup = lookup;
 	master_source_unlock(ap->entry);
 
-	/* If we don't need to create directories then there's no use
-	 * reading the map. We just need to test that the map is valid
-	 * for the fail cases to function correctly and to cache the
-	 * lookup handle.
-	 *
-	 * We always need to whole map for direct mounts in order to
-	 * mount the triggers.
-	 */
-	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT)
-		return NSS_STATUS_SUCCESS;
-
 	if (!map->stale)
 		return NSS_STATUS_SUCCESS;
 
@@ -298,14 +295,20 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 
 	status = lookup->lookup_read_map(ap, age, lookup->context);
 
-	map->stale = 0;
+	if (status != NSS_STATUS_SUCCESS)
+		map->stale = 0;
 
 	/*
 	 * For maps that don't support enumeration return success
 	 * and do whatever we must to have autofs function with an
 	 * empty map entry cache.
+	 *
+	 * For indirect maps that use the browse option, when the
+	 * server is unavailable continue as best we can with
+	 * whatever we have in the cache, if anything.
 	 */
-	if (status == NSS_STATUS_UNKNOWN)
+	if (status == NSS_STATUS_UNKNOWN ||
+	   (ap->type == LKP_INDIRECT && status == NSS_STATUS_UNAVAIL))
 		return NSS_STATUS_SUCCESS;
 
 	return status;
@@ -410,6 +413,7 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 	tmap.instance = map->instance;
 	tmap.recurse = map->recurse;
 	tmap.depth = map->depth;
+	tmap.stale = map->stale;
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
@@ -597,7 +601,7 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 		cache_readlock(mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
-			if (*me->key == '*')
+			if (!strcmp(me->key, "*"))
 				goto next;
 
 			if (*me->key == '/') {
@@ -685,7 +689,7 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 	char *type, *format;
 
 	if (stat(map->argv[0], &st) == -1) {
-		warn(ap->logopt, "file map not found");
+		debug(ap->logopt, "file map not found");
 		return NSS_STATUS_NOTFOUND;
 	}
 
@@ -804,6 +808,49 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 	return result;
 }
 
+static void update_negative_cache(struct autofs_point *ap, struct map_source *source, const char *name)
+{
+	struct master_mapent *entry = ap->entry;
+	struct map_source *map;
+	struct mapent *me;
+
+	/* Don't update negative cache for included maps */ 
+	if (source && source->depth)
+		return;
+
+	/* Have we recorded the lookup fail for negative caching? */
+	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
+	if (me)
+		/*
+		 *  Already exists in the cache, the mount fail updates
+		 *  will update negative timeout status.
+		 */
+		cache_unlock(me->mc);
+	else {
+		/* Notify only once after fail */
+		logmsg("key \"%s\" not found in map source(s).", name);
+
+		/* Doesn't exist in any source, just add it somewhere */
+		if (source)
+			map = source;
+		else
+			map = entry->maps;
+		if (map) {
+			time_t now = time(NULL);
+			int rv = CHE_FAIL;
+
+			cache_writelock(map->mc);
+			rv = cache_update(map->mc, map, name, NULL, now);
+			if (rv != CHE_FAIL) {
+				me = cache_lookup_distinct(map->mc, name);
+				me->status = now + ap->negative_timeout;
+			}
+			cache_unlock(map->mc);
+		}
+	}
+	return;
+}
+
 int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const char *name, int name_len)
 {
 	struct master_mapent *entry = ap->entry;
@@ -907,8 +954,13 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		send_map_update_request(ap);
 	pthread_cleanup_pop(1);
 
-	if (result == NSS_STATUS_NOTFOUND)
-		error(ap->logopt, "key \"%s\" not found in map.", name);
+	/*
+	 * The last source lookup will return NSS_STATUS_NOTFOUND if the
+	 * map exits and the key has not been found but the map may also
+	 * not exist in which case the key is also not found.
+	 */
+	if (result == NSS_STATUS_NOTFOUND || result == NSS_STATUS_UNAVAIL)
+		update_negative_cache(ap, source, name);
 
 	return !result;
 }
@@ -955,25 +1007,128 @@ static char *make_fullpath(const char *root, const char *key)
 		if (l > KEY_MAX_LEN)
 			return NULL;
 		path = malloc(l);
+		if (!path)
+			return NULL;
 		strcpy(path, key);
 	} else {
 		l = strlen(key) + 1 + strlen(root) + 1;
 		if (l > KEY_MAX_LEN)
 			return NULL;
 		path = malloc(l);
+		if (!path)
+			return NULL;
 		sprintf(path, "%s/%s", root, key);
 	}
 	return path;
 }
 
+void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, time_t age)
+{
+	struct mapent *me, *this;
+	char *path;
+	int status = CHE_FAIL;
+
+	me = cache_enumerate(mc, NULL);
+	while (me) {
+		struct mapent *valid;
+		char *key = NULL, *next_key = NULL;
+
+		if (me->age >= age) {
+			/*
+			 * Reset time of last fail for valid map entries to
+			 * force entry update and subsequent mount retry.
+			 */
+			me->status = 0;
+			me = cache_enumerate(mc, me);
+			continue;
+		}
+
+		key = strdup(me->key);
+		me = cache_enumerate(mc, me);
+		if (!key || !strcmp(key, "*")) {
+			if (key)
+				free(key);
+			continue;
+		}
+
+		path = make_fullpath(ap->path, key);
+		if (!path) {
+			warn(ap->logopt, "can't malloc storage for path");
+			free(key);
+			continue;
+		}
+
+		/*
+		 * If this key has another valid entry we want to prune it,
+		 * even if it's a mount, as the valid entry will take the
+		 * mount if it is a direct mount or it's just a stale indirect
+		 * cache entry.
+		 */
+		valid = lookup_source_valid_mapent(ap, key, LKP_DISTINCT);
+		if (valid && valid->mc == mc) {
+			 /*
+			  * We've found a map entry that has been removed from
+			  * the current cache so it isn't really valid.
+			  */
+			cache_unlock(valid->mc);
+			valid = NULL;
+		}
+		if (!valid &&
+		    is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
+			debug(ap->logopt,
+			      "prune check posponed, %s mounted", path);
+			free(key);
+			free(path);
+			continue;
+		}
+		if (valid)
+			cache_unlock(valid->mc);
+
+		if (me)
+			next_key = strdup(me->key);
+
+		cache_unlock(mc);
+
+		cache_writelock(mc);
+		this = cache_lookup_distinct(mc, key);
+		if (!this) {
+			cache_unlock(mc);
+			goto next;
+		}
+
+		if (valid)
+			cache_delete(mc, key);
+		else if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
+			status = CHE_FAIL;
+			if (this->ioctlfd == -1)
+				status = cache_delete(mc, key);
+			if (status != CHE_FAIL) {
+				if (ap->type == LKP_INDIRECT) {
+					if (ap->flags & MOUNT_FLAG_GHOST)
+						rmdir_path(ap, path, ap->dev);
+				} else
+					rmdir_path(ap, path, this->dev);
+			}
+		}
+		cache_unlock(mc);
+
+next:
+		cache_readlock(mc);
+		if (next_key) {
+			me = cache_lookup_distinct(mc, next_key);
+			free(next_key);
+		}
+		free(key);
+		free(path);
+	}
+
+	return;
+}
+
 int lookup_prune_cache(struct autofs_point *ap, time_t age)
 {
 	struct master_mapent *entry = ap->entry;
 	struct map_source *map;
-	struct mapent_cache *mc;
-	struct mapent *me, *this;
-	char *path;
-	int status = CHE_FAIL;
 
 	pthread_cleanup_push(master_source_lock_cleanup, entry);
 	master_source_readlock(entry);
@@ -985,86 +1140,11 @@ int lookup_prune_cache(struct autofs_point *ap, time_t age)
 			map = map->next;
 			continue;
 		}
-		mc = map->mc;
-		pthread_cleanup_push(cache_lock_cleanup, mc);
-		cache_readlock(mc);
-		me = cache_enumerate(mc, NULL);
-		while (me) {
-			char *key = NULL, *next_key = NULL;
-
-			if (me->age >= age) {
-				me = cache_enumerate(mc, me);
-				continue;
-			}
-
-			key = strdup(me->key);
-			me = cache_enumerate(mc, me);
-			if (!key || *key == '*') {
-				if (key)
-					free(key);
-				continue;
-			}
-
-			path = make_fullpath(ap->path, key);
-			if (!path) {
-				warn(ap->logopt,
-				     "can't malloc storage for path");
-				free(key);
-				continue;
-			}
-
-			if (is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
-				debug(ap->logopt,
-				      "prune check posponed, %s mounted", path);
-				free(key);
-				free(path);
-				continue;
-			}
-
-			if (me)
-				next_key = strdup(me->key);
-
-			cache_unlock(mc);
-
-			cache_writelock(mc);
-			this = cache_lookup_distinct(mc, key);
-			if (!this) {
-				cache_unlock(mc);
-				free(key);
-				if (next_key)
-					free(next_key);
-				free(path);
-				goto next;
-			}
-
-			if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
-				status = CHE_FAIL;
-				if (this->ioctlfd == -1)
-					status = cache_delete(mc, key);
-				if (status != CHE_FAIL) {
-					if (ap->type == LKP_INDIRECT) {
-						if (ap->flags & MOUNT_FLAG_GHOST)
-							rmdir_path(ap, path, ap->dev);
-					} else
-						rmdir_path(ap, path, this->dev);
-				}
-			}
-			cache_unlock(mc);
-
-			if (!next_key) {
-				free(key);
-				free(path);
-				cache_readlock(mc);
-				continue;
-			}
-next:
-			cache_readlock(mc);
-			me = cache_lookup_distinct(mc, next_key);
-			free(key);
-			free(path);
-			free(next_key);
-		}
+		pthread_cleanup_push(cache_lock_cleanup, map->mc);
+		cache_readlock(map->mc);
+		lookup_prune_one_cache(ap, map->mc, age);
 		pthread_cleanup_pop(1);
+		clear_stale_instances(map);
 		map->stale = 0;
 		map = map->next;
 	}
@@ -1082,7 +1162,6 @@ struct mapent *lookup_source_valid_mapent(struct autofs_point *ap, const char *k
 	struct mapent_cache *mc;
 	struct mapent *me = NULL;
 
-	master_source_readlock(entry);
 	map = entry->maps;
 	while (map) {
 		/*
@@ -1105,7 +1184,6 @@ struct mapent *lookup_source_valid_mapent(struct autofs_point *ap, const char *k
 		cache_unlock(mc);
 		map = map->next;
 	}
-	master_source_unlock(entry);
 
 	return me;
 }
