@@ -1,9 +1,21 @@
-#include <alloca.h>
+/* ----------------------------------------------------------------------- *
+ *
+ *  cat_path.c - boundary aware buffer management routines
+ *
+ *   Copyright 2002-2003 Ian Kent <raven@themaw.net> - All Rights Reserved
+ *
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
+ *   USA; either version 2 of the License, or (at your option) any later
+ *   version; incorporated herein by reference.
+ *
+ * ----------------------------------------------------------------------- */
+
 #include <string.h>
 #include <limits.h>
-#include <syslog.h>
 #include <ctype.h>
-
+#include "automount.h"
 /*
  * sum = "dir/base" with attention to buffer overflows, and multiple
  * slashes at the joint are avoided.
@@ -44,16 +56,15 @@ int cat_path(char *buf, size_t len, const char *dir, const char *base)
 	return 1;
 }
 
-int _strlen(const char *str, int max)
+size_t _strlen(const char *str, size_t max)
 {
-	char *s = (char *) str;
+	const char *s = str;
+	size_t len = 0;
 
-	while (isprint(*s++) && max--) ;
+	while (*s++ && len < max)
+		len++;
 
-	if (max < 0)
-		return 0;
-	
-	return s - str - 1;
+	return len;
 }
 
 /* 
@@ -65,7 +76,7 @@ int ncat_path(char *buf, size_t len,
 	      const char *dir, const char *base, size_t blen)
 {
 	char name[PATH_MAX+1];
-	int alen = _strlen(base, blen);
+	size_t alen = _strlen(base, blen);
 
 	if (blen > PATH_MAX || !alen)
 		return 0;
