@@ -185,10 +185,8 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	 * reading the map. We always need to read the whole map for
 	 * direct mounts in order to mount the triggers.
 	 */
-	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT) {
-		debug(ap->logopt, "map read not needed, so not done");
+	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT)
 		return NSS_STATUS_SUCCESS;
-	}
 
 	mc = source->mc;
 
@@ -243,19 +241,7 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 		if (*key == '+')
 			continue;
 
-		if (!(source->flags & MAP_FLAG_FORMAT_AMD))
-			s_key = sanitize_path(key, len, ap->type, ap->logopt);
-		else {
-			if (!strcmp(key, "/defaults")) {
-				mapent = ENTRY_VAL(this, 1);
-				cache_writelock(mc);
-				cache_update(mc, source, key, mapent, age);
-				cache_unlock(mc);
-				continue;
-			}
-			/* Don't fail on "/" in key => type == 0 */
-			s_key = sanitize_path(key, len, 0, ap->logopt);
-		}
+		s_key = sanitize_path(key, len, ap->type, ap->logopt);
 		if (!s_key)
 			continue;
 
@@ -279,10 +265,10 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 }
 
 static int lookup_one(struct autofs_point *ap,
-		      struct map_source *source,
 		      const char *key, int key_len,
 		      struct lookup_context *ctxt)
 {
+	struct map_source *source;
 	struct mapent_cache *mc;
 	char *tablename;
 	nis_result *result;
@@ -292,6 +278,10 @@ static int lookup_one(struct autofs_point *ap,
 	int ret, cur_state;
 	char buf[MAX_ERR_BUF];
 
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
 	mc = source->mc;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -334,67 +324,9 @@ static int lookup_one(struct autofs_point *ap,
 	return ret;
 }
 
-static int match_key(struct autofs_point *ap,
-		     struct map_source *source,
-		     const char *key, int key_len,
-		     struct lookup_context *ctxt)
-{
-	unsigned int is_amd_format = source->flags & MAP_FLAG_FORMAT_AMD;
-	char buf[MAX_ERR_BUF];
-	char *lkp_key;
-	char *prefix;
-	int ret;
-
-	ret = lookup_one(ap, source, key, key_len, ctxt);
-	if (ret < 0)
-		return ret;
-	if (ret == CHE_OK || ret == CHE_UPDATED || is_amd_format)
-		return ret;
-
-	lkp_key = strdup(key);
-	if (!lkp_key) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "strdup: %s", estr);
-		return CHE_FAIL;
-	}
-
-	ret = CHE_MISSING;
-
-	/*
-	 * Now strip successive directory components and try a
-	 * match against map entries ending with a wildcard and
-	 * finally try the wilcard entry itself.
-	 */
-	while ((prefix = strrchr(lkp_key, '/'))) {
-		char *match;
-		size_t len;
-		*prefix = '\0';
-		len = strlen(lkp_key) + 3;
-		match = malloc(len);
-		if (!match) {
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt, MODPREFIX "malloc: %s", estr);
-			ret = CHE_FAIL;
-			goto done;
-		}
-		len--;
-		strcpy(match, lkp_key);
-		strcat(match, "/*");
-		ret = lookup_one(ap, source, match, len, ctxt);
-		free(match);
-		if (ret < 0)
-			goto done;
-		if (ret == CHE_OK || ret == CHE_UPDATED)
-			goto done;
-	}
-done:
-	free(lkp_key);
-	return ret;
-}
-
-static int lookup_wild(struct autofs_point *ap,
-		       struct map_source *source, struct lookup_context *ctxt)
+static int lookup_wild(struct autofs_point *ap, struct lookup_context *ctxt)
 {
+	struct map_source *source;
 	struct mapent_cache *mc;
 	char *tablename;
 	nis_result *result;
@@ -404,6 +336,10 @@ static int lookup_wild(struct autofs_point *ap,
 	int ret, cur_state;
 	char buf[MAX_ERR_BUF];
 
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
 	mc = source->mc;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -444,84 +380,28 @@ static int lookup_wild(struct autofs_point *ap,
 	return ret;
 }
 
-static int lookup_amd_defaults(struct autofs_point *ap,
-			       struct map_source *source,
-			       struct lookup_context *ctxt)
-{
-	struct mapent_cache *mc = source->mc;
-	char *tablename;
-	nis_result *result;
-	nis_object *this;
-	char *mapent;
-	char buf[MAX_ERR_BUF];
-	int cur_state;
-	int ret;
-
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	tablename = malloc(9 + strlen(ctxt->mapname) +
-			   strlen(ctxt->domainname) + 20);
-	if (!tablename) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		pthread_setcancelstate(cur_state, NULL);
-		return CHE_FAIL;
-	}
-	sprintf(tablename, "[key=/defaults],%s.org_dir.%s",
-		ctxt->mapname, ctxt->domainname);
-
-	result = nis_list(tablename, FOLLOW_PATH | FOLLOW_LINKS, NULL, NULL);
-	if (result->status != NIS_SUCCESS && result->status != NIS_S_SUCCESS) {
-		nis_error rs = result->status;
-		nis_freeresult(result);
-		free(tablename);
-		pthread_setcancelstate(cur_state, NULL);
-		if (rs == NIS_NOTFOUND ||
-		    rs == NIS_S_NOTFOUND ||
-		    rs == NIS_PARTIAL)
-			return CHE_MISSING;
-
-		return -rs;
-	}
-
-	this = NIS_RES_OBJECT(result);
-	mapent = ENTRY_VAL(this, 1);
-
-	cache_writelock(mc);
-	ret = cache_update(mc, source, "/defaults", mapent, time(NULL));
-	cache_unlock(mc);
-
-	nis_freeresult(result);
-	free(tablename);
-	pthread_setcancelstate(cur_state, NULL);
-
-	return ret;
-}
-
 static int check_map_indirect(struct autofs_point *ap,
-			      struct map_source *source,
 			      char *key, int key_len,
 			      struct lookup_context *ctxt)
 {
-	unsigned int is_amd_format = source->flags & MAP_FLAG_FORMAT_AMD;
+	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *me, *exists;
 	time_t now = time(NULL);
 	time_t t_last_read;
 	int ret = 0;
 
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
 	mc = source->mc;
 
-	if (is_amd_format) {
-		/* Check for a /defaults entry to update the map source */
-		if (lookup_amd_defaults(ap, source, ctxt) == CHE_FAIL) {
-			warn(ap->logopt, MODPREFIX
-			     "error getting /defaults from map %s",
-			     ctxt->mapname);
-		}
-	}
+	master_source_current_wait(ap->entry);
+	ap->entry->current = source;
 
 	/* check map and if change is detected re-read map */
-	ret = match_key(ap, source, key, key_len, ctxt);
+	ret = lookup_one(ap, key, key_len, ctxt);
 	if (ret == CHE_FAIL)
 		return NSS_STATUS_NOTFOUND;
 
@@ -530,16 +410,9 @@ static int check_map_indirect(struct autofs_point *ap,
 		 * If the server is down and the entry exists in the cache
 		 * and belongs to this map return success and use the entry.
 		 */
-		cache_readlock(mc);
-		if (source->flags & MAP_FLAG_FORMAT_AMD)
-			exists = match_cached_key(ap, MODPREFIX, source, key);
-		else
-			exists = cache_lookup(mc, key);
-		if (exists && exists->source == source) {
-			cache_unlock(mc);
+		exists = cache_lookup(mc, key);
+		if (exists && exists->source == source)
 			return NSS_STATUS_SUCCESS;
-		}
-		cache_unlock(mc);
 
 		warn(ap->logopt,
 		     MODPREFIX "lookup for %s failed: %s",
@@ -558,10 +431,6 @@ static int check_map_indirect(struct autofs_point *ap,
 		}
 		me = cache_lookup_next(mc, me);
 	}
-	if (is_amd_format)
-		exists = match_cached_key(ap, MODPREFIX, source, key);
-	else
-		exists = cache_lookup_distinct(mc, key);
 	exists = cache_lookup_distinct(mc, key);
 	/* Not found in the map but found in the cache */
 	if (exists && exists->source == source && ret & CHE_MISSING) {
@@ -581,7 +450,10 @@ static int check_map_indirect(struct autofs_point *ap,
 		int wild = CHE_MISSING;
 		struct mapent *we;
 
-		wild = lookup_wild(ap, source, ctxt);
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+
+		wild = lookup_wild(ap, ctxt);
 		/*
 		 * Check for map change and update as needed for
 		 * following cache lookup.
@@ -618,11 +490,9 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	struct mapent_cache *mc;
 	char key[KEY_MAX_LEN + 1];
 	int key_len;
-	char *lkp_key;
 	char *mapent = NULL;
 	int mapent_len;
 	struct mapent *me;
-	char buf[MAX_ERR_BUF];
 	int status;
 	int ret = 1;
 
@@ -634,18 +504,9 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 
 	debug(ap->logopt, MODPREFIX "looking up %s", name);
 
-	if (!(source->flags & MAP_FLAG_FORMAT_AMD)) {
-		key_len = snprintf(key, KEY_MAX_LEN + 1, "%s", name);
-		if (key_len > KEY_MAX_LEN)
-			return NSS_STATUS_NOTFOUND;
-	} else {
-		key_len = expandamdent(name, NULL, NULL);
-		if (key_len > KEY_MAX_LEN)
-			return NSS_STATUS_NOTFOUND;
-		memset(key, 0, KEY_MAX_LEN + 1);
-		expandamdent(name, key, NULL);
-		debug(ap->logopt, MODPREFIX "expanded key: \"%s\"", key);
-	}
+	key_len = snprintf(key, KEY_MAX_LEN + 1, "%s", name);
+	if (key_len > KEY_MAX_LEN)
+		return NSS_STATUS_NOTFOUND;
 
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, key, LKP_DISTINCT);
@@ -664,10 +525,8 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 				cache_writelock(smc);
 				sme = cache_lookup_distinct(smc, key);
 				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent) {
-					if (cache_pop_mapent(sme) == CHE_FAIL)
-						cache_delete(smc, key);
-				}
+				if (sme && !sme->mapent)
+					cache_delete(smc, key);
 				cache_unlock(smc);
 			}
 		}
@@ -679,65 +538,29 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	 * we never know about it.
 	 */
 	if (ap->type == LKP_INDIRECT && *key != '/') {
+		char *lkp_key;
+
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, key);
 		if (me && me->multi)
 			lkp_key = strdup(me->multi->key);
-		else if (!ap->pref)
+		else
 			lkp_key = strdup(key);
-		else {
-			lkp_key = malloc(strlen(ap->pref) + strlen(key) + 1);
-			if (lkp_key) {
-				strcpy(lkp_key, ap->pref);
-				strcat(lkp_key, key);
-			}
-		}
 		cache_unlock(mc);
 
-		if (!lkp_key) {
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt, MODPREFIX "malloc: %s", estr);
+		if (!lkp_key)
 			return NSS_STATUS_UNKNOWN;
-		}
 
-		status = check_map_indirect(ap, source,
-					    lkp_key, strlen(lkp_key), ctxt);
-		free(lkp_key);
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+
+		status = check_map_indirect(ap, lkp_key, strlen(lkp_key), ctxt);
 		if (status)
 			return status;
 	}
 
-	/*
-	 * We can't take the writelock for direct mounts. If we're
-	 * starting up or trying to re-connect to an existing direct
-	 * mount we could be iterating through the map entries with
-	 * the readlock held. But we don't need to update the cache
-	 * when we're starting up so just take the readlock in that
-	 * case.
-	 */
-	if (ap->flags & MOUNT_FLAG_REMOUNT)
-		cache_readlock(mc);
-	else
-		cache_writelock(mc);
-
-	if (!ap->pref)
-		lkp_key = strdup(key);
-	else {
-		lkp_key = malloc(strlen(ap->pref) + strlen(key) + 1);
-		if (lkp_key) {
-			strcpy(lkp_key, ap->pref);
-			strcat(lkp_key, key);
-		}
-	}
-
-	if (!lkp_key) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "malloc: %s", estr);
-		cache_unlock(mc);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	me = match_cached_key(ap, MODPREFIX, source, lkp_key);
+	cache_readlock(mc);
+	me = cache_lookup(mc, key);
 	/* Stale mapent => check for entry in alternate source or wildcard */
 	if (me && !me->mapent) {
 		while ((me = cache_lookup_key_next(me)))
@@ -746,26 +569,12 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		if (!me)
 			me = cache_lookup_distinct(mc, "*");
 	}
-	if (me && me->mapent) {
-		/*
-		 * If this is a lookup add wildcard match for later validation
-		 * checks and negative cache lookups.
-		 */
-		if (!(ap->flags & MOUNT_FLAG_REMOUNT) &&
-		    ap->type == LKP_INDIRECT && *me->key == '*') {
-			ret = cache_update(mc, source, key, me->mapent, me->age);
-			if (!(ret & (CHE_OK | CHE_UPDATED)))
-				me = NULL;
-		}
-		if (me && (me->source == source || *me->key == '/')) {
-			mapent_len = strlen(me->mapent);
-			mapent = malloc(mapent_len + 1);
-			if (mapent)
-				strcpy(mapent, me->mapent);
-		}
+	if (me && me->mapent && (me->source == source || *me->key == '/')) {
+		mapent_len = strlen(me->mapent);
+		mapent = malloc(mapent_len + 1);
+		strcpy(mapent, me->mapent);
 	}
 	cache_unlock(mc);
-	free(lkp_key);
 
 	if (!mapent)
 		return NSS_STATUS_TRYAGAIN;
@@ -780,22 +589,16 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		time_t now = time(NULL);
 		int rv = CHE_OK;
 
-		free(mapent);
-
-		/* Don't update negative cache when re-connecting */
-		if (ap->flags & MOUNT_FLAG_REMOUNT)
-			return NSS_STATUS_TRYAGAIN;
 		cache_writelock(mc);
 		me = cache_lookup_distinct(mc, key);
-		if (me)
-			rv = cache_push_mapent(me, NULL);
-		else
+		if (!me)
 			rv = cache_update(mc, source, key, NULL, now);
 		if (rv != CHE_FAIL) {
 			me = cache_lookup_distinct(mc, key);
 			me->status = time(NULL) + ap->negative_timeout;
 		}
 		cache_unlock(mc);
+		free(mapent);
 		return NSS_STATUS_TRYAGAIN;
 	}
 	free(mapent);
